package jp.sourceforge.acerola3d.a3.bvh;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import jp.sourceforge.acerola3d.a3.bvh.analysis.*;
import jp.sourceforge.acerola3d.a3.bvh.lexer.*;
import jp.sourceforge.acerola3d.a3.bvh.node.*;
import jp.sourceforge.acerola3d.a3.bvh.parser.*;


class ParserBVH extends DepthFirstAdapter {
    int frameCount;
    double frameTime;
    String rootBone;
    HashMap<String,Bone> bones = new HashMap<String,Bone>();

    //BVHをパース中に使用するテンポラリデータが入ってる変数
    Tmp t;

    public ParserBVH(URL url) throws Exception {
        InputStream is = url.openStream();
        Reader reader = new InputStreamReader(is);
        Lexer lexer = new Lexer(new PushbackReader(reader));
        Parser parser = new Parser(lexer);
        jp.sourceforge.acerola3d.a3.bvh.node.Node ast = parser.parse();

        t = new Tmp();
        ast.apply(this); // <- BVHをパースしてtの中にデータを構築する．
        convert(); // <- tの中の生のBVHデータをTransform3Dに変換
        t = null;
    }

    public double getMotionLength() {
        return frameCount*frameTime;
    }

    public double getDefaultFrameTime() {
        return frameTime;
    }

    public String getRootBone() {
        return rootBone;
    }

    public String[] getChildBones(String b) {
        Bone bone = bones.get(b);
        return bone.children.clone();
    }

    public Transform3D getTransform3D(String bone,double time) {
        Bone b = bones.get(bone);
        int frame = (int)(time/frameTime);
        frame = (frame<0)?0:frame;
        frame = (frame>=frameCount)?frameCount-1:frame;
        return new Transform3D(b.transforms[frame]);
    }

    //最終的にBoneの情報を保存しておくためのクラス
    class Bone {
        String name;
        String children[];
        Transform3D transforms[];
    }

    //パースする時にBVHのBlockのデータをそのままの形で保存しておくための
    //クラス．Blockは最終的にはBoneに変換される．
    class Block {
        String name;
        ArrayList<String> children;
        double  offsetX;
        double  offsetY;
        double  offsetZ;
        Channel x_position;
        Channel y_position;
        Channel z_position;
        Channel z_rotation;
        Channel x_rotation;
        Channel y_rotation;
        Channel x_scale;
        Channel y_scale;
        Channel z_scale;
    }

    //Channelのデータを保存するためのクラス
    //Channelは一つのBlock対して複数含まれる．
    class Channel {
        double data[];
    }

    //パース中に一時的に必要となるデータをまとめて
    //保存するためのクラス．少しでもメモリを節約して
    //みようというもがきクラス．
    class Tmp {
        int frameCount;
        double frameTime;
        ArrayList<Block> blockQueue;
        Stack<Block> blockStack;
        ArrayList<Channel> channelQueue;

        Block rootBlock;
        Block parentBlock;
        Block currentBlock;

        Tmp() {
            blockQueue = new ArrayList<Block>();
            blockStack = new Stack<Block>();
            channelQueue = new ArrayList<Channel>();
        }
        void channelsInit() {
            Iterator<Channel> i = channelQueue.iterator();
            while (i.hasNext()) {
                i.next().data = new double[frameCount];
            }
        }
    }

//------------------------------------------------------------
//以下BVHをパースするためのメソッド
//------------------------------------------------------------

    //以下文法解析の部分
    //tの中にあるparentBlock,currentBlock,blockStackあたりの同期は，
    //各ブロックのcaseA???のメソッドに入った瞬間にやることとする．
    //また，今回は混乱しそうなので，inA???,outA???のメソッドは
    //使わないことにする．
    public void caseABvh(ABvh node){
        node.getHierarchyPart().apply(this);
        node.getMotionPart().apply(this);
    }

    public void caseAHierarchyPart(AHierarchyPart node) {
        node.getRootBlock().apply(this);
    }

    public void caseARootBlock(ARootBlock node) {
        t.rootBlock = new Block();
        t.currentBlock = t.rootBlock;
        t.blockQueue.add(t.currentBlock);

        t.currentBlock.name = node.getIdentifier().getText();
        t.currentBlock.children = new ArrayList<String>();

        node.getOffsetLine().apply(this);
        node.getChannelsLine().apply(this);

        Object temp[] = node.getSuperBlock().toArray();
        for(int i = 0; i < temp.length; i++) {
            ((PSuperBlock) temp[i]).apply(this);
        }
    }

    public void caseAOffsetLine(AOffsetLine node) {
        PNumber x_number = node.getXOffset();
        PNumber y_number = node.getYOffset();
        PNumber z_number = node.getZOffset();

        t.currentBlock.offsetX = pNumber2Double(x_number);
        t.currentBlock.offsetY = pNumber2Double(y_number);
        t.currentBlock.offsetZ = pNumber2Double(z_number);
    }

    public void caseAChannelsLine(AChannelsLine node) {
        TInteger integer = node.getInteger();
        int channelsCount = Integer.parseInt(integer.getText());

        Object temp[] = node.getChannelType().toArray();
        if (channelsCount != temp.length) {
            System.out.println("ActionGenerator.caseAChannelsBlock()");
            System.out.println("pars error!!!");
        }

        for (int i=0;i<temp.length;i++) {
            Channel c = new Channel();
            t.channelQueue.add(c);
            if (temp[i] instanceof AXPositionChannelType) {
                t.currentBlock.x_position = c;
            } else if (temp[i] instanceof AYPositionChannelType) {
                t.currentBlock.y_position = c;
            } else if (temp[i] instanceof AZPositionChannelType) {
                t.currentBlock.z_position = c;
            } else if (temp[i] instanceof AXRotationChannelType) {
                t.currentBlock.x_rotation = c;
            } else if (temp[i] instanceof AYRotationChannelType) {
                t.currentBlock.y_rotation = c;
            } else if (temp[i] instanceof AZRotationChannelType) {
                t.currentBlock.z_rotation = c;
            } else if (temp[i] instanceof AXScaleChannelType) {
                t.currentBlock.x_scale = c;
            } else if (temp[i] instanceof AYScaleChannelType) {
                t.currentBlock.y_scale = c;
            } else if (temp[i] instanceof AZScaleChannelType) {
                t.currentBlock.z_scale = c;
            }
        }
    }

    public void caseAJointBlock(AJointBlock node) {
        t.blockStack.push(t.parentBlock);
        t.parentBlock = t.currentBlock;
        t.currentBlock = new Block();
        t.blockQueue.add(t.currentBlock);

        t.currentBlock.name = node.getIdentifier().getText();
        t.currentBlock.children = new ArrayList<String>();

        t.parentBlock.children.add(t.currentBlock.name);

        node.getOffsetLine().apply(this);
        node.getChannelsLine().apply(this);

        Object temp[] = node.getSuperBlock().toArray();
        for(int i = 0; i < temp.length; i++)
                ((PSuperBlock) temp[i]).apply(this);

        t.currentBlock = t.parentBlock;
        t.parentBlock = t.blockStack.pop();
    }

    public void caseAEndSiteBlock(AEndSiteBlock node) {
        t.blockStack.push(t.parentBlock);
        t.parentBlock = t.currentBlock;
        t.currentBlock = new Block();

        t.currentBlock.name = null;
        node.getOffsetLine().apply(this);

        t.currentBlock = t.parentBlock;
        t.parentBlock = t.blockStack.pop();
    }

    public void caseAMotionPart(AMotionPart node) {
        node.getFramesLine().apply(this);
        node.getFrameTimeLine().apply(this);

        t.channelsInit();

        Object temp[] = node.getNumber().toArray();
        if (temp.length != (t.channelQueue.size()*t.frameCount)) {
            System.out.println("ActionGenerator.caseAMotionPart().");
            System.out.println("parse error!!!");
            System.out.print(temp.length);
            System.out.print(" and "+t.channelQueue.size()*t.frameCount);
            System.out.println(" do not match.");
            return;
        }

        int tempCount = 0;
        for (int i=0;i<t.frameCount;i++) {
            for (int j=0;j<t.channelQueue.size();j++) {
                PNumber pn = (PNumber)temp[tempCount++];
                double d = pNumber2Double(pn);
                t.channelQueue.get(j).data[i] = d;
            }
        }
    }

    public void caseAFramesLine(AFramesLine node) {
        TInteger integer = node.getInteger();
        String is = integer.getText();
        t.frameCount = Integer.parseInt(is);
    }

    public void caseAFrameTimeLine(AFrameTimeLine node) {
        PNumber num = node.getNumber();
        t.frameTime = pNumber2Double(num);
    }

//----------

    protected double pNumber2Double(PNumber num) {
        if (num instanceof ARealNumberNumber) {
            TRealNumber rn = ((ARealNumberNumber)num).getRealNumber();
            String rns = rn.getText();
            return Double.parseDouble(rns);
        } else {
            TInteger integer = ((AIntegerNumber)num).getInteger();
            int iii = Integer.parseInt(integer.getText());
            return (double)iii;
        }
    }

//------------------------------------------------------------
//BVHの生データをTransform3Dに変換する．
//------------------------------------------------------------

    void convert() {
        frameCount = t.frameCount;
        frameTime = t.frameTime;
        rootBone = t.rootBlock.name;

        //たぶん，ループの中でnewすると無駄が多くなるのでここで宣言しておく
        Transform3D t1 = new Transform3D();
        Transform3D t2 = new Transform3D();
        double xyz[] = new double[3];
        Vector3d v = new Vector3d();

        Iterator<Block> i = t.blockQueue.iterator();
        while (i.hasNext()) {
            Block block = i.next();
            Bone bone = new Bone();
            bone.name = block.name;
            bone.children = block.children.toArray(new String[0]);
            bone.transforms = new Transform3D[frameCount];
            for (int j=0;j<frameCount;j++) {
                Transform3D t0 = new Transform3D();
                t0.setTranslation(new Vector3d(block.offsetX,block.offsetY,block.offsetZ));
                if (block.z_rotation!=null) {
                    t1.rotZ(block.z_rotation.data[j]/360.0*2.0*Math.PI);
                }
                if (block.x_rotation!=null) {
                    t2.rotX(block.x_rotation.data[j]/360.0*2.0*Math.PI);
                    t1.mul(t2);
                }
                if (block.y_rotation!=null) {
                    t2.rotY(block.y_rotation.data[j]/360.0*2.0*Math.PI);
                    t1.mul(t2);
                }
                xyz[0] = xyz[1] = xyz[2] = 0.0;
                if (block.x_position!=null) {
                    xyz[0] = block.x_position.data[j];
                }
                if (block.y_position!=null) {
                    xyz[1] = block.y_position.data[j];
                }
                if (block.z_position!=null) {
                    xyz[2] = block.z_position.data[j];
                }
                v.set(xyz);
                t1.setTranslation(v);
                xyz[0] = xyz[1] = xyz[2] = 1.0;
                if (block.x_scale!=null) {
                    xyz[0] = block.x_scale.data[j];
                }
                if (block.y_scale!=null) {
                    xyz[1] = block.y_scale.data[j];
                }
                if (block.z_scale!=null) {
                    xyz[2] = block.z_scale.data[j];
                }
                v.set(xyz);
                t1.setScale(v);
                t0.mul(t1);
                bone.transforms[j] = t0;
            }
            bones.put(bone.name,bone);
        }
    }
}
